/*
 * Network features to implement:
 *
 * - Work Offline mode (no passive network access)
 * - Get login status
 * - Log in
 */


var Network = {
  get offline () {
    return getIOService().offline;
  },


  set offline (val) {
    getIOService().offline = val;
    return val;
  },


  get loggedIn () {
    if (this.offline)
      return false;

    var authMgr = getAuthManager();

    // Check if already auth'd, callback true if so
    var rv = { domain: {}, username: {}, password: {} };

    try {
      // Throws an exception if not auth'd
      authMgr.getAuthIdentity(Cx.AUTH_SCHEME, Cx.PUBLISH_SERVER, Cx.AUTH_PORT,
                              Cx.AUTH_TYPE, Cx.AUTH_REALM, Cx.AUTH_PATH,
                              rv.domain, rv.username, rv.password);
      return true;
    }
    catch (ex) {
      return false;
    }
  },


  get username () {
    if (this.offline)
      return null;

    var authMgr = getAuthManager();

    // Check if already auth'd, callback true if so
    var rv = { domain: {}, username: {}, password: {} };

    try {
      // Throws an exception if not auth'd
      authMgr.getAuthIdentity(Cx.AUTH_SCHEME, Cx.PUBLISH_SERVER, Cx.AUTH_PORT,
                              Cx.AUTH_TYPE, Cx.AUTH_REALM, Cx.AUTH_PATH,
                              rv.domain, rv.username, rv.password);
      return rv.username.value;
    }
    catch (ex) {
      return null;
    }
  },


  login: function login (service, callback, reattempt) {
    var cxsvc = getCeltxService();
    var observer = {
      onLogin: function (succeeded) {}
    };
    cxsvc.login(service, observer, reattempt, window);
    return;

    var authMgr = getAuthManager();

    if (this.loggedIn) {
      // XXX maybe test-auth with current credentials anyway?
      callback(true);
      return;
    }

    // XXX deadman timer?
    var request = new XMLHttpRequest();
    request.onload  = function () {
      var authMgr = getAuthManager();
      if (request.status == 401) {
        // Not Authorized
        dump("401 not authorized\n");
        authMgr.clearAll();
        // Retry
        Network.login(window, service, callback, true);
      }
      else if (request.status != 200) {
        dump("error: status: " + request.status + "\n");
        authMgr.clearAll();
        if (callback)
          callback(false);
      }
      else {
        getObserverService().notifyObservers(this,
          "celtx:login-status-changed", "loggedin");
        if (callback)
          callback(true);
      }
    };
    request.onerror = function () {
      if (callback)
        callback(false);
    };
    request.open("GET", Cx.AUTH_URL);
    var channel = request.channel;
    var prompt = new CeltxAuthPrompt(service, channel.notificationCallbacks,
                                     request);
    if (reattempt)
      prompt.reattempt = true;
    channel.notificationCallbacks = prompt;
    request.send(null);
  },


  logout: function () {
    var cxsvc = Components.classes["@celtx.com/celtx-service;1"]
      .getService(Components.interfaces.nsICeltxService);
    cxsvc.logout();
    return;

    var authMgr = getAuthManager();
    authMgr.clearAll();
    getObserverService().notifyObservers(this,
      "celtx:login-status-changed", "loggedin");
  }
};
